import bart_tf

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

def real_from_complex_weights(wgh):
    import tensorflow as tf

    shp = wgh.shape

    filter_depth, filter_height, filter_width, in_channels, out_channels, tmp = shp
    size = [filter_depth, filter_height, filter_width, in_channels, out_channels, 1]
    
    rwgh=tf.slice(wgh, begin=[0,0,0,0,0,0], size = size)
    iwgh=tf.slice(wgh, begin=[0,0,0,0,0,1], size = size)

    rwgh = tf.reshape(rwgh, [filter_depth, filter_height, filter_width, in_channels, 1, out_channels, 1])
    iwgh = tf.reshape(iwgh, [filter_depth, filter_height, filter_width, in_channels, 1, out_channels, 1])

    wgh = tf.concat([tf.concat([rwgh, iwgh], 6), tf.concat([-iwgh, rwgh], 6)], 4)

    return tf.reshape(wgh, [filter_depth, filter_height, filter_width, 2 * in_channels, 2 * out_channels])


def tf1_generate_resnet(path, model):

    tf.reset_default_graph()

    batch_size = None
    image_shape = [1, 32, 32]
    img_channel = 1

    img = tf.placeholder(tf.float32, shape=[batch_size] + [1] * 12 + image_shape+[2], name='input_0')
    img_t = tf.reshape(img, [-1] + image_shape + [img_channel * 2])

    conv_0 = tf.placeholder(tf.float32, shape=[1, 3, 3, 1, 8, 2], name='input_1')
    conv_i = tf.placeholder(tf.float32, shape=[1, 1, 3, 3, 8, 8, 2], name='input_2')
    conv_n = tf.placeholder(tf.float32, shape=[1, 3, 3, 8, 1, 2], name='input_3')

    conv_is = tf.reshape(conv_i, [1, 3, 3, 8, 8, 2])

    wghts = []
    wghts.append(real_from_complex_weights(conv_0))
    wghts.append(real_from_complex_weights(conv_is))
    wghts.append(real_from_complex_weights(conv_n))

    out = img_t
    for wgh in wghts:
        out = tf.nn.conv3d(out, wgh, [1] * 5, "SAME")
        if wgh != wghts[-1]:
            out = tf.nn.relu(out)
        
    out = out + img_t
    out = tf.reshape(out, tf.shape(img), name='output_0')

    bart_tf.tf1_export_graph(path, name=model)

tf1_generate_resnet("./", "tf1_resnet")






import sys
import traceback
import os

try:
    with open(sys.argv[1], 'r') as fd:
        for line in fd.readlines():
            exec(line)               
except:
    exc_info = sys.exc_info()
    traceback.print_exception(*exc_info)
    print('Exception occurred while executing line: ', line)
    sys.exit(1)




#!/usr/bin/python
# -- coding: utf-8 --

"""
Copyright 2022. Uecker Lab, University Medical Center Goettingen.

Authors:
2022 Martin Schilling (martin.schilling@med.uni-goettingen.de)
2022 Nick Scholand (scholand@tugraz.at)

DESCRIPTION :
This script takes an SVG file as an input, analyses the paths of objects,
which can consist of horizontal, vertical, diagonal and cubic spline
transformations, splits these transformations up into cubic Hermite splines
and creates a CFL file for use with the BART phantom command line tool.
"""

import os
import numpy as np
import argparse

import sys
if 'BART_TOOLBOX_PATH' in os.environ and os.path.exists(os.environ['BART_TOOLBOX_PATH']):
    sys.path.insert(0, os.path.join(os.environ['BART_TOOLBOX_PATH'], 'python'))
elif 'TOOLBOX_PATH' in os.environ and os.path.exists(os.environ['TOOLBOX_PATH']):
    sys.path.insert(0, os.path.join(os.environ['TOOLBOX_PATH'], 'python'))
else:
    raise RuntimeError("BART_TOOLBOX_PATH is not set correctly!")
import cfl

DBLEVEL = 0


def read_svg(svg_input, scale_flag=True):
	"""
	Reads a given svg file to extract parameters of paths.

	:param str svg_input: File path to input svg.
	:param bool scale_coords: Boolean for scaling coordinates
	:returns: List of lists for paths. Element: [object_id, color, transforms]
	:rtype: list
	"""
	paths_list = []
	prev_key_list, points_list = [], []
	readout = False

	with open (svg_input, 'rt', encoding="utf8", errors='ignore') as input:
		for line in input:
			if "<path" in line:
				readout = True
				object_id="000"
				color = "000000"
				prev_keys=[]
				points=[]

			if readout:
				if "style=" in line:
					if "stroke:#" in line:
						color = line.split("stroke:#")[1][:6]

					if "fill:#" in line:
						color = line.split("fill:#")[1][:6]

				if "     d=" in line:
					prev_keys, points = analyse_d_string(line.split('"')[1])

				if "id=" in line:
					object_id = line.split('"')[1]

				# end of parameters
				if "/>" in line:
					readout = False
					prev_key_list.append(prev_keys)
					points_list.append(points)
					paths_list.append([object_id, color])
	input.close()

	if scale_flag:
		scale_coords(points_list, center=[0,0], norm=1.8)

	for num,(k,p) in enumerate(zip(prev_key_list,points_list)):
		if 0 != len(k):
			transforms = get_transforms(k,p)
			paths_list[num].append(transforms)
		else:
			paths_list[num].append([])

	return paths_list

def scale_coords(coords_list, center=[187.5, 125], norm=350):
	"""
	Scale coordinates to a given center with a maximal norm.

	:param list coords_list: List of lists of coordinates. Each list belongs to a series of control points.
	:param list center: Coordinates of new center
	:param int norm: Maximal value for width and height
	"""
	x_min = coords_list[0][0][0]
	x_max = coords_list[0][0][0]
	y_min = coords_list[0][0][1]
	y_max = coords_list[0][0][1]

	# determine maximal and minimal x- and y-values
	for cs in coords_list:
		for c in cs:
			x_max = c[0] if c[0] > x_max else x_max
			x_min = c[0] if c[0] < x_min else x_min
			y_max = c[1] if c[1] > y_max else y_max
			y_min = c[1] if c[1] < y_min else y_min

	# transfer values to new center
	x_trans = (x_max + x_min) / 2
	y_trans = (y_max + y_min) / 2
	# normalization factor of coordinates as ratio of norm and max([width,height])
	norm_factor = norm / max([np.abs(x_max-x_min), np.abs(y_max-y_min)])

	for cs in coords_list:
		for c in cs:
			c[0] = (c[0] - x_trans) * norm_factor + center[0]
			c[1] = (c[1] - y_trans) * norm_factor + center[1]

def try_float(string):
	# Function for trying a string for conversion to float.
	try:
		f = float(string)
		return True
	except ValueError:
		return False

def analyse_d_string(d_string):
	"""
	Analyse string of 'd' argument of path. The function returns the signal transform parameters
	and a list of coordinates of the control points.

	:param str d_string: Complete string contained in the 'd' parameter of a path.
	:returns: tuple(transform_keys, coordinates)
		WHERE
		list transform_keys is list of lower case signal characters for transformations
		list coordinates is list of absolute coordinates of control points
	"""
	content = d_string.split()
	prev_key = None
	points = []
	transf_keys = []
	cspline = []
	count = 0
	x_origin, y_origin = 0, 0
	for num,section in enumerate(content):
		if len(section.split(",")) > 1 or try_float(section):
			# keys before coordinates signal new transformation,
			# lower case for relative, upper case for absolute coordinates
			special_keys = ['c','C','m','M', 'l', 'L']

			# deal with exception, that 'm'/'M' keys may be followed by diagonal
			# transformation ('l' key) without explicit key
			if len(section.split(",")) > 1 and prev_key not in special_keys:
				prev_key = "l"

			if "c" == prev_key or "C" == prev_key:

				count += 1

				if 3 == count:
					cspline.append([cspline[1][0]+cspline[1][0]-cspline[0][0],cspline[1][1]+cspline[1][1]-cspline[0][1]])
					if "c" == prev_key:
						# relative reference point
						x_origin += float(content[num].split(",")[0])
						y_origin += float(content[num].split(",")[1])
					if "C" == prev_key:
						# absolute reference point
						x_origin = float(content[num].split(",")[0])
						y_origin = float(content[num].split(",")[1])

					# append intermediate control points
					points.append([cspline[0][0], cspline[0][1]])
					transf_keys.append(prev_key)
					points.append([cspline[1][0], cspline[1][1]])
					transf_keys.append(prev_key)

					count = 0
					cspline = []
				else:
					if "c" == prev_key:
						cspline.append([x_origin+float(content[num].split(",")[0]), y_origin+float(content[num].split(",")[1])])
					if "C" == prev_key:
						cspline.append([float(content[num].split(",")[0]), float(content[num].split(",")[1])])
			else:
				count = 0

			# start of path
			if "m" == prev_key or "M" == prev_key:
				x_origin = float(content[num].split(",")[0])
				y_origin = float(content[num].split(",")[1])

			# horizontal transformation
			if "h" == prev_key:
				x_origin += float(content[num])
			if "H" == prev_key:
				x_origin = float(content[num])

			# vertical transformation
			if "v" == prev_key:
				y_origin += float(content[num])
			if "V" == prev_key:
				y_origin = float(content[num])

			# diagonal transformation
			if "l" == prev_key:
				x_origin += float(content[num].split(",")[0])
				y_origin += float(content[num].split(",")[1])
			if "L" == prev_key:
				x_origin = float(content[num].split(",")[0])
				y_origin = float(content[num].split(",")[1])

			if 0 == count:
				points.append([x_origin, y_origin])
				transf_keys.append(prev_key.lower())
				if 'M' == prev_key:
					prev_key = 'L'
				if 'm' == prev_key:
					prev_key = 'l'
		else:
			prev_key = section

	return transf_keys, points

def controlpoints2cspline(bezier_points):
	"""
	Translate four input control points into a cubic Hermite spline format suitable for BART.

	:param list bezier_points: List of four control points in format [p1,p2,p3,p4] with p_i=[x_i,y_i]
	:returns: Parameters for cubic Hermite spline [x_parameters, y_parameters]
	:rtype: list
	"""
	bezier_cspline = [[1,-3,0,0],[0,3,0,0],[0,0,0,-3],[0,0,1,3]]
	bezier_x = [p[0] for p in bezier_points]
	bezier_y = [p[1] for p in bezier_points]
	bezier = [bezier_x, bezier_y]
	cspline = [[0,0]  for i in range(4)]

	for num, c in enumerate(bezier):
		for i in range(4):
			for j in range(4):
				cspline[i][num] += bezier_cspline[j][i] * bezier[num][j]

	cspline_x = [p[0] for p in cspline]
	cspline_y = [p[1] for p in cspline]

	return [cspline_x, cspline_y]

def get_transforms(keys, points):
	"""
	Create separate transformations from given lists of keys and coordinates.
	The transformations have the form [[x_transforms],[y_transforms]] in the cubic Hermite spline format.

	:param list keys: List of signal characters for path transformations [key1, key2, ...]
	:param list points: List of coordinates [[x1,y1], [x2,y2], ...]
	:returns: Transformations in cubic Hermite spline format
	:rtype: list
	"""
	transforms = []

	for num,(k,p) in enumerate(zip(keys,points)):

		if 'h' == k:
			transforms.append([[points[num-1][0],0.,p[0],-0.],[p[1],0.,p[1],-0.]])

		if 'v' == k:
			transforms.append([[p[0],0.,p[0],-0.],[points[num-1][1],0.,p[1],-0.]])

		if 'l' == k:
			transforms.append([[points[num-1][0],0.,p[0],-0.],[points[num-1][1],0.,p[1],-0.]])

		if num+1 < len(points) and 'c' == k:
			# non-trivial B-spline
			if 'c' == keys[num-1] and 'c' == keys[num+1]:
				keys[num+1] = None
				transforms.append(controlpoints2cspline(points[num-2:num+2]))

			# trivial B-spline
			elif 'c' != keys[num-1] and 'c' != keys[num+1]:
				transforms.append([[points[num-1][0],0.,p[0],-0.],[points[num-1][1],0.,p[1],-0.]])

	return transforms

def format_transforms(transforms, object_id, filename, output_file):
	"""
	Format transforms for insertion into /bart/src/geom/logo.c

	:param list transform: List of transformations in cubic Hermite spline format
	:param list object_id: List of object ids for indexing transformations
	:param str filename: Name of struct
	:param str output_file: File path to output text file
	"""
	total_transforms = sum([len(t) for t in transforms])
	with open (output_file, 'w', encoding="utf8", errors='ignore') as output:
		output.write("//Replace in bart/src/geom/logo.c > bart_logo and adjust bart/src/geom/logo.h\n\n")
		output.write("const double "+filename+"["+str(total_transforms)+"][2][4] = {\n")
		for num, transform in enumerate(transforms):
			output.write("\t//"+str(object_id[num])+"\n")
			for enum,t in enumerate(transform):
				x_string = str(t[0][0])+", "+str(t[0][1])+", "+str(t[0][2])+", "+str(t[0][3])
				y_string = str(t[1][0])+", "+str(t[1][1])+", "+str(t[1][2])+", "+str(t[1][3])
				# current implementation in BART, likely to change to x_string, y_string in the future
				output.write("\t{ { "+y_string+" }, { "+x_string+" } },\n")
		output.write("};\n")

def transform2polystruct(transforms, id_color, output_file):
	"""
	Create a polystruct for a given set of transformations and append it to output file.
	Can replace code in bart/src/simu/phantom.c > calc_bart

	:param list transforms: List of transformations in [[x_transforms],[y_transforms]] format
	:param list id_color: List of fill colors of individual objects
	:param str output_file: File path to output text file
	"""
	total_transforms = sum([len(t) for t in transforms])

	with open (output_file, 'a', encoding="utf8", errors='ignore') as output:
		output.write("\tint N = "+str(total_transforms)+";\n")
		output.write("\tdouble points[N * 11][2];\n")
		output.write("\n")
		output.write("\tstruct poly poly = {\n\t\tkspace,\n\t\tcoeff,\n\t\tpopts->large_sens,\n\t\t"+str(len(transforms))+",\n\t\t&(struct poly1[]){\n")
		array_position = 0
		for num, transform in enumerate(transforms):
			output.write("\t\t\t{ "+str(len(transform)*11)+" , "+str(id_color[num])+", ARRAY_SLICE(points, "+str(array_position*11)+", "+str((array_position+len(transform))*11) +") },\n")
			array_position += len(transform)
		output.write("\t\t}\n")
		output.write("\t};")
	output.close()

def assign_color_id(colors):
	"""
	Extract color IDs from hex colors

	:param list colors: List of strings representing the objects colors

	:returns: List of Integers representing the objects colors as integer (> 0 !) IDs
	:rtype: list
	"""

	color_values, color_counts = np.unique(colors, return_counts=True)

	id_color = [list(color_values).index(i)+1 for i in colors]

	return id_color

# Save geometry data in numpy array
# 	coord -> [segment, cp_set:[x,y], cp_coord] with control points (cp)
#	meta -> [path index, number of segments, color of path]
def save2cfl(new_transforms, new_colors, cfl_output):

	coord = []
	meta = []
	ind_path = 0

	for sub_array in new_transforms:

		ind_seg = 0

		for path in sub_array:

			path_array = np.array(path)
			coord.append(path_array)

			ind_seg += 1

		meta.append(np.array([ind_path, ind_seg, new_colors[ind_path]]))
		ind_path += 1

	coord = np.array(coord)
	meta = np.array(meta)
	
	if (2 <= DBLEVEL):
		print("Coord Dims:")
		print(np.shape(coord))
		print("Meta Dims:")
		print(np.shape(meta))
		print("Meta:")
		print(meta)

	cfl.writemulticfl(cfl_output, np.array([coord, meta], dtype=object))


def main(svg_input, text_output, output):
	"""
	Extract parameters of paths from SVG file and write code block into txt file,
	which is suitable for bart/src/simu/shepplogan.c > calc_bart and bart/src/geom/logo.c.

	:param str svg_input: File path to input SVG file
	:param str cfl: File path to output cfl file. Default: <svg_name>.{cfl,hdr}
	"""

	if (text_output):
		text_filename = output+".txt"

	path_objects = read_svg(svg_input)

	object_ids = [obj[0] for obj in path_objects]
	colors = [obj[1] for obj in path_objects]
	transforms = [obj[2] for obj in path_objects]

	# Sort paths by color (=: grey value in provided SVG file)

	id_color = assign_color_id(colors)

	new_colors = sorted(id_color)
	new_ids = [id for color, id in sorted(zip(id_color,object_ids))]
	new_transforms = [trans for color, trans in sorted(zip(id_color,transforms))]

	color_values, color_counts = np.unique(new_colors, return_counts=True)

	if (2 <= DBLEVEL):
		print("Distribution of colors:")
		print("Value:\t", color_values)
		print("Number:\t", color_counts)


	save2cfl(new_transforms, new_colors, output)


	if (1 <= DBLEVEL):
		print("Created files:")
		print(output+".{cfl,hdr}")


	if (text_output):

		format_transforms(new_transforms, new_ids, output, text_filename)

		transform2polystruct(new_transforms, new_colors, text_filename)

		if (1 <= DBLEVEL):
			print(output+".txt")

if __name__ == "__main__":

	parser = argparse.ArgumentParser(
		description="Script to extract control points of cubic Hermite splines from SVG file to CFL format.")

	parser.add_argument('input', type=str, help="Input SVG file")
	parser.add_argument('output', type=str, help="Output CFL filename")
	parser.add_argument('-d', '--db', default=-1, type=int, help="Specify debug value for additional information  [default: 0]")

	# Internal option for more complicated objects with multi component paths (example: BRAIN geometry)
	# Requires manual tuning and is therefore hidden for simplicity
	parser.add_argument('-t', action='store_true', help=argparse.SUPPRESS)

	args = parser.parse_args()

	if ("BART_DEBUG_LEVEL" in os.environ):

		if (-1 != args.db):
			print("A local BART_DEBUG_LEVEL variable exists! It will be overwritten by -d input!\n")

		DBLEVEL = int(os.environ["BART_DEBUG_LEVEL"])

	elif ("DEBUG_LEVEL" in os.environ):
		DBLEVEL = int(os.environ["DEBUG_LEVEL"])

	if (-1 != args.db):
		DBLEVEL = args.db

	main(args.input, args.t, args.output)



# Copyright 2016. The Regents of the University of California.
# All rights reserved. Use of this source code is governed by 
# a BSD-style license which can be found in the LICENSE file.
#
# Authors: 
# 2016 Siddharth Iyer <sid8795@gmail.com>
# 2018 Soumick Chatterjee <soumick.chatterjee@ovgu.de> , WSL Support

import subprocess as sp
import tempfile as tmp
import os
import sys
from pathlib import Path
import threading

if __spec__.parent:
    from . import cfl
else:  # 'traditional' copy-paste bart.py
    import cfl

if os.name == 'nt':
    if __spec__.parent:
        from .wslsupport import PathCorrection
    else:
        from wslsupport import PathCorrection

isWASM = sys.platform == 'emscripten'


def bart(nargout, cmd, *args, **kwargs):
    if isWASM:
        print("Please use await bart.bart2!", file=sys.stderr)
        raise RuntimeError("Synchronous bart not available in wasm")

    prep = bart_prepare(nargout, cmd, *args, **kwargs)
    if not prep:
        return

    ERR, stdout, stderr = execute_cmd(prep['shell_cmd'])

    # store error code, stdout and stderr in function attributes for outside access
    # this makes it possible to access these variables from outside the function (e.g "print(bart.ERR)")
    bart.ERR, bart.stdout, bart.stderr = ERR, stdout, stderr

    return bart_postprocess(nargout, ERR, prep['infiles'], prep['infiles_kw'], prep['outfiles'])


async def bart2(nargout, cmd, *args, **kwargs):
    if not isWASM:
        print("Please use synchronous bart.bart!", file=sys.stderr)
        raise RuntimeError("Asynchronous bart is only available in wasm")

    prep = bart_prepare(nargout, cmd, *args, **kwargs)
    if not prep:
        return

    ERR, stdout, stderr = await run_wasm_cmd(**prep)

    # store error code, stdout and stderr in function attributes for outside access
    # this makes it possible to access these variables from outside the function (e.g "print(bart.ERR)")
    bart.ERR, bart.stdout, bart.stderr = ERR, stdout, stderr

    return bart_postprocess(nargout, ERR, prep['infiles'], prep['infiles_kw'], prep['outfiles'])


def bart_prepare(nargout, cmd, *args, **kwargs):
    if type(nargout) is not int or nargout < 0:
        print("Usage: bart(<nargout>, <command>, <arguments...>)")
        return

    bart_path = os.environ.get('BART_TOOLBOX_PATH')

    isWSL = False

    if not isWASM and not bart_path:
        if os.path.isfile('/usr/local/bin/bart'):
            bart_path = '/usr/local/bin'
        elif os.path.isfile('/usr/bin/bart'):
            bart_path = '/usr/bin'
        else:
            bartstatus = os.system('wsl bart version -V')
            if bartstatus==0:
                bart_path = '/usr/bin'
                isWSL = True
            else:
                raise Exception('Environment variable BART_TOOLBOX_PATH is not set.')

    name = tmp.NamedTemporaryFile().name

    nargin = len(args)
    infiles = [name + 'in' + str(idx) + '.ra' for idx in range(nargin)]

    for idx in range(nargin):
        cfl.writecfl(infiles[idx], args[idx])

    args_kw = [("--" if len(kw)>1 else "-") + kw for kw in kwargs]
    infiles_kw = [name + 'in' + kw + '.ra' for kw in kwargs]
    for idx, kw in enumerate(kwargs):
        cfl.writecfl(infiles_kw[idx], kwargs[kw])

    outfiles = [name + 'out' + str(idx) + '.ra' for idx in range(nargout)]

    cmd = cmd.split(" ")

    if os.name =='nt':
        if isWSL:
            #For WSL and modify paths
            infiles = [PathCorrection(item) for item in infiles]
            infiles_kw = [PathCorrection(item) for item in infiles_kw]
            outfiles = [PathCorrection(item) for item in outfiles]
            cmd = [PathCorrection(item) for item in cmd]
            args_infiles_kw = [item for pair in zip(args_kw, infiles_kw) for item in pair]
            shell_cmd = ['wsl', 'bart', *cmd, *args_infiles_kw, *infiles, *outfiles]
        else:
            #For cygwin use bash and modify paths
            infiles = [item.replace(os.path.sep, '/') for item in infiles]
            infiles_kw = [item.replace(os.path.sep, '/') for item in infiles_kw]
            outfiles = [item.replace(os.path.sep, '/') for item in outfiles]
            cmd = [item.replace(os.path.sep, '/') for item in cmd]
            args_infiles_kw = [item for pair in zip(args_kw, infiles_kw) for item in pair]
            shell_cmd = ['bash.exe', '--login',  '-c', os.path.join(bart_path, 'bart'), *cmd, *args_infiles_kw, *infiles, *outfiles]
            #TODO: Test with cygwin, this is just translation from matlab code

        assert(not isWASM)

    else:
        args_infiles_kw = [item for pair in zip(args_kw, infiles_kw) for item in pair]
        shell_cmd = [os.path.join(bart_path, 'bart') if not isWASM else 'bart', *cmd, *args_infiles_kw, *infiles, *outfiles]

    return dict(shell_cmd=shell_cmd, infiles=infiles, infiles_kw=infiles_kw, outfiles=outfiles)


def bart_postprocess(nargout, ERR, infiles, infiles_kw, outfiles):
    for elm in infiles + infiles_kw:
        if elm.endswith('.ra'):
            Path(elm).unlink(missing_ok=True)
        else:
            Path(elm + '.hdr').unlink(missing_ok=True)
            Path(elm + '.cfl').unlink(missing_ok=True)

    output = []
    for idx in range(nargout):
        elm = outfiles[idx]
        try:
            if not ERR:
                output.append(cfl.readcfl(elm))
        finally:
            if elm.endswith('.ra'):
                Path(elm).unlink(missing_ok=True)
            else:
                Path(elm + '.hdr').unlink(missing_ok=True)
                Path(elm + '.cfl').unlink(missing_ok=True)

    if ERR:
        print(f"Command exited with error code {ERR}.")
        return

    if nargout == 0:
        return
    elif nargout == 1:
        return output[0]
    else:
        return output


def execute_cmd(cmd):
    """
    Execute a command in a shell.
    Print and catch the output (stdout + stderr) in real time.
    """

    # remove empty strings from cmd
    cmd = [item for item in cmd if len(item)]

    # execute cmd
    proc = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, universal_newlines=True)

    def read_stream(stream, buffer):
        for line in iter(stream.readline, ""):
            buffer.append(line)
            print(line, end="")
        stream.close()

    stdout_lines, stderr_lines = [], []

    t_out = threading.Thread(target=read_stream, args=(proc.stdout, stdout_lines))
    t_err = threading.Thread(target=read_stream, args=(proc.stderr, stderr_lines))

    t_out.start()
    t_err.start()

    t_out.join()
    t_err.join()

    errcode = proc.wait()
    return errcode, "".join(stdout_lines), "".join(stderr_lines)



wasm_bart_ok = False

async def get_wasm_cfl(name):
    await wasm_async_call(f"get_cfl('{name}')")

async def put_wasm_cfl(name):
    await wasm_async_call(f"send_cfl('{name}')")

async def rm_bart_cfl(name):
    await wasm_async_call(f"rm_cfl('{name}')")

async def wasm_load_bart():
    global wasm_bart_ok
    await wasm_async_call("reload_bart()")
    wasm_bart_ok = True

async def run_wasm_cmd(shell_cmd, infiles, infiles_kw, outfiles):
    global wasm_bart_ok
    try:
        if not wasm_bart_ok:
            await wasm_load_bart()

        for f in infiles + infiles_kw:
            await put_wasm_cfl(f)

        non_empty_cmd = [x for x in shell_cmd if len(shell_cmd) > 0]

        result = await wasm_async_call("bart_cmd('" + ' '.join(non_empty_cmd) + "')")
        ERR, stdout, stderr = result['ret'], result['stdout'], result['stderr']

        if stdout is not None and len(stdout.strip()) > 0:
            print(stdout)

        if stderr is not None and len(stderr.strip()) > 0:
            print(stderr, file=sys.stderr)

        if not 0 == ERR:
            print(f"Function exited with {ERR}", file=sys.stderr)

        for f in outfiles:
            await get_wasm_cfl(f)

        for f in infiles + infiles_kw + outfiles:
            await rm_bart_cfl(f)

        return ERR, stdout, stderr

    except Exception as e:
        print("Exception in bart worker calls occurred:")
        print(e)
        wasm_bart_ok = False
        raise e

async def wasm_async_call(cmd):
    # synchronous function would be nice
    # but this seems impossible: https://github.com/pyodide/pyodide/issues/3932
    #loop = asyncio.get_event_loop()
    #task = pyodide.code.run_js(cmd)
    #return loop.run_until_complete(asyncio.wait([task]))

    ret = (await pyodide.code.run_js(cmd)).to_py()
    if 0 != ret[0]:
        raise Exception(f"Error in JS call: {ret[1]}")

    return ret[1]


if isWASM:
    import pyodide, pyodide_js, js

    # Export pyodide to webworker namespace:
    pyodide.code.run_js("var pyodide;")
    js.pyodide = pyodide_js

    # load BART:
    pyodide.code.run_js("""
                        if ('undefined' == (typeof window)) {
                            importScripts("bart_base.js");

                        } else {

                            script = document.createElement('script');
                            script.type = 'text/javascript';
                            script.async = true;
                            script.src = "bart_base.js";
                            document.body.appendChild(script);
                        }
                        """)



import string
import os

def PathCorrection(inData):
	outData=inData
	for i in string.ascii_lowercase: #Replace drive letters with /mnt/
		outData=outData.replace(i+':','/mnt/'+i) #if drive letter is supplied in lowercase
		outData=outData.replace(i.upper()+':','/mnt/'+i) #if drive letter is supplied as uppercase
	outData=outData.replace(os.path.sep, '/') #Change windows filesep to linux filesep

	return outData



import argparse
import numpy as np
import h5py

from bart import cfl

def convert(infile, outfile):
    x = h5py.File(infile, 'r')
    dat = np.array(x['dataset']['data'])['data']

    cmplx = np.zeros((len(dat[0])//2, len(dat)), dtype=np.complex64)

    for i,acq in enumerate(dat):
        cmplx[:,i].real = acq[::2]
        cmplx[:,i].imag = acq[1::2]

    cfl.writecfl(outfile, cmplx)


parser = argparse.ArgumentParser()
parser.add_argument("infile")
parser.add_argument("outfile")
args = parser.parse_args()

convert(args.infile, args.outfile)



# Copyright 2023. Christian Tönnes.
# All rights reserved. Use of this source code is governed by
# a BSD-style license which can be found in the LICENSE file.
# 
# Authors:
# 2023 Christian Tönnes <christian.toennes@keks.li>
#

# pip install libsass watchdog
from http.server import SimpleHTTPRequestHandler, HTTPServer
import mimetypes

class RequestHandler(SimpleHTTPRequestHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory="./wwwroot", **kwargs)

    if not mimetypes.inited:
        mimetypes.init() # try to read system mime.types
    extensions_map = mimetypes.types_map.copy()
    extensions_map.update({
        '': 'application/octet-stream', # Default
        '.py': 'text/plain',
        '.c': 'text/plain',
        '.h': 'text/plain',
        '.js': 'application/javascript',
        })    

handler_class = RequestHandler
httpd = HTTPServer(('', 8000), handler_class)

httpd.serve_forever()


import bart_tf

def real_from_complex_weights(wgh):
    import tensorflow as tf

    shp = wgh.shape

    filter_depth, filter_height, filter_width, in_channels, out_channels, tmp = shp
    size = [filter_depth, filter_height, filter_width, in_channels, out_channels, 1]
    
    rwgh=tf.slice(wgh, begin=[0,0,0,0,0,0], size = size)
    iwgh=tf.slice(wgh, begin=[0,0,0,0,0,1], size = size)

    rwgh = tf.reshape(rwgh, [filter_depth, filter_height, filter_width, in_channels, 1, out_channels, 1])
    iwgh = tf.reshape(iwgh, [filter_depth, filter_height, filter_width, in_channels, 1, out_channels, 1])

    wgh = tf.concat([tf.concat([rwgh, iwgh], 6), tf.concat([-iwgh, rwgh], 6)], 4)

    return tf.reshape(wgh, [filter_depth, filter_height, filter_width, 2 * in_channels, 2 * out_channels])


def tf2_generate_resnet(path, model):

    import tensorflow as tf
    import numpy as np

    class ComplexConv3D(tf.Module):
        def __init__(self, filters, kernel_size, dummy_dim = False):
            super().__init__()
            # filters: 64, kernel_size: 3, stride: 1
            self.filters = filters
            self.kernel_size = kernel_size
            self.is_built = False
            self.dummy_dim = dummy_dim

        def __call__(self, input):

            if not(self.is_built):
                if self.dummy_dim:
                    shp = [1] + list(self.kernel_size) + [input.shape[-2], self.filters, 2]
                else:
                    shp = list(self.kernel_size) + [input.shape[-2], self.filters, 2]

                scale = np.sqrt(1 / (np.prod(self.kernel_size) * self.filters + input.shape[-2]))
                self.conv_weight = tf.Variable(tf.random.normal(shp, stddev=scale), name='w')
                self.is_built = True

            conv = self.conv_weight
            if self.dummy_dim:
                conv = tf.reshape(conv, conv.shape[1:])
            conv = real_from_complex_weights(conv)

            shp = tf.shape(input)
            shp = tf.concat([shp[:-2], [2 *shp[-2]]], 0)

            tmp = tf.reshape(input, shp)
            tmp = tf.nn.conv3d(tmp, conv, [1] * 5, "SAME")

            shp = tf.shape(tmp)
            shp = tf.concat([shp[:-1], [shp[-1] // 2, 2]], 0)

            return tf.reshape(tmp, shp)
    
    class ResBlock(tf.Module):
        def __init__(self):
            super().__init__()
            self.conv1 = ComplexConv3D(8, (1, 3, 3))
            self.conv2 = ComplexConv3D(8, (1, 3, 3), dummy_dim=True)
            self.conv3 = ComplexConv3D(1, (1, 3, 3))

        def __call__(self, input):

            shp = tf.shape(input)
            shp = tf.concat([shp[:-1], tf.constant([1]), shp[-1:]], 0)

            out = tf.reshape(input, shp)
            out = self.conv1(out)
            out = tf.nn.relu(out)
            out = self.conv2(out)
            out = tf.nn.relu(out)
            out = self.conv3(out)
            out = input + tf.reshape(out, tf.shape(input))
            return out

    bart_tf.tf2_export_module(ResBlock(), [32, 32, 1], path+"/"+model, trace_complex=False)

tf2_generate_resnet("./", "tf2_resnet")





import os
import sys

if 'BART_TOOLBOX_PATH' in os.environ and os.path.exists(os.environ['BART_TOOLBOX_PATH']):
	sys.path.append(os.path.join(os.environ['BART_TOOLBOX_PATH'], 'python'))
else:
	raise RuntimeError("BART_TOOLBOX_PATH is not set correctly!")


from bart import bart
import cfl



#!/usr/bin/env python3

import cfl
import sys

# from http://stackoverflow.com/questions/5574702/how-to-print-to-stderr-in-python
def errprint(*args, **kwargs):
	    print(*args, file=sys.stderr, **kwargs)

def main(out_name, in_name):
	input = cfl.readmulticfl(in_name)
	#cfl.writecfl(input, out_name)
	cfl.writemulticfl(out_name, input)
	return 0

if __name__ == '__main__':

	if len(sys.argv) != 3:
		errprint('Usage:', sys.argv[0], '<input> <output>')

	exit(main(sys.argv[2], sys.argv[1]))



# Copyright 2022. Uecker Lab. University Center Göttingen.
# All rights reserved. Use of this source code is governed by
# a BSD-style license which can be found in the LICENSE file.
#
# Authors:
# Moritz Blumenthal

import os
import numpy as np
import cfl

import tensorflow as tf2

try:
    import tensorflow.compat.v1 as tf1
except ImportError:
    import tensorflow as tf1
    pass

def tf2_export_module(model, dims, path, trace_complex=True):
    class BartWrapper(tf2.Module):
    
        def __init__(self, model, dims, vars_as_input = True, name=None):
        
            super(BartWrapper, self).__init__(name=name)

            self.model = model
            self.trace_complex = trace_complex

            self.dims_bart = [1] * 16
            self.dims_tf =   [1] * (len(dims) + 1)

            if not(trace_complex):

                self.dims_bart = self.dims_bart + [2]
                self.dims_tf = self.dims_tf + [2]
            
                for i in range(len(dims)):
                    self.dims_bart[len(self.dims_bart) - 2 - i] = dims[i]
                    self.dims_tf[len(self.dims_tf) - 2 - i] = dims[i]
                
                self.model(np.zeros(self.dims_tf, np.float32)) #run model ones to initialize weights

            else :
                for i in range(len(dims)):
                    self.dims_bart[len(self.dims_bart) - 2 - i] = dims[i]
                    self.dims_tf[len(self.dims_tf) - 2 - i] = dims[i]

                self.model(np.zeros(self.dims_tf, np.complex64)) #run model ones to initialize weights

            self.dims_tf[0] = -1
            self.dims_bart[0] = -1

            self.trace_complex = trace_complex
        
            if vars_as_input:
                self.vars = model.variables
            else:
                self.vars = []

            self.vars_rtoc = [] # variables for which a 0 imaginary part is stacked
            for var in self.vars:
                self.vars_rtoc.append(2 != var.shape[-1])

            self.sig = {}

            self.add_concrete_function()

        @tf2.function
        def __call__(self, input, weights, grad_in):

            for i in range(len(weights)):
            
                wgh = weights[i]

                if (self.vars_rtoc)[i]:    
                    slc = [ slice(None, None, None) ]* len(wgh.shape)
                    slc[-1] = 0
                    wgh=wgh[tuple(slc)]
                
                self.model.variables[i].assign(wgh)


            with tf2.GradientTape(persistent=True) as g:
                g.watch(input)

                print("Tracing TensorFlow model with dims: {}".format(input))

                res = tf2.reshape(input, self.dims_tf)

                outr = self.model(res)

                out = tf2.reshape(outr, self.dims_bart)

            result = {}
        
            result["output_0"] = out
            result["grad_0_0"] = g.gradient(out, input, grad_in)

            for i, input in enumerate(self.model.variables, 1):
                result["grad_{}_0".format(i)] = g.gradient(out, input, grad_in)
                
                if self.vars_rtoc[i - 1]:
                    tmp = result["grad_{}_0".format(i)]
                    result["grad_{}_0".format(i)] = tf2.stack([tmp, tf2.zeros_like(tmp)], axis = len(tmp.shape))
        
            return result


        def add_concrete_function(self, name=None):

            dims = self.dims_bart.copy()
            dims[0] = None

            if (self.trace_complex):
                signature_input = tf2.TensorSpec(shape=dims, dtype=tf2.complex64, name="input_0")
                signature_grad_ys = tf2.TensorSpec(shape=dims, dtype=tf2.complex64, name="grad_ys_0")
            else:
                signature_input = tf2.TensorSpec(shape=dims, dtype=tf2.float32, name="input_0")
                signature_grad_ys = tf2.TensorSpec(shape=dims, dtype=tf2.float32, name="grad_ys_0")

            signature_weight = []
            for i, var in enumerate(self.model.variables, 1):
                if self.vars_rtoc[i - 1]:
                    signature_weight.append(tf2.TensorSpec(shape=list(var.shape)+[2], dtype=tf2.float32, name="input_{}".format(i)))
                else:
                    signature_weight.append(tf2.TensorSpec(shape=var.shape, dtype=tf2.float32, name="input_{}".format(i)))

            if name is None:
                name = "serving_default"
            self.sig[name] = self.__call__.get_concrete_function(signature_input, signature_weight, signature_grad_ys)


        def save_variables(self, path):
            
            weights = []
            for i, var in enumerate(self.variables):

                if (self.vars_rtoc[i]):
                    weights.append(var.numpy().astype(np.complex64)) 
                else:
                    weights.append(np.empty(var.shape[:-1], dtype=np.complex64))
                    slc = [ slice(None, None, None) ] * len(var.shape)
                    slc[-1] = 0
                    weights[-1].real = var.numpy()[tuple(slc)]
                    slc[-1] = 1
                    weights[-1].imag = var.numpy()[tuple(slc)]
                    
                    if 0 == len(weights[-1].shape):
                        weights[-1] = weights[-1].reshape([1]) 
            
                weights[-1] = np.transpose(weights[-1])
            
            if (0 < len(weights)):
                cfl.writemulticfl(path, weights)



        def save(self, path):

            tf2.saved_model.save(self, path, signatures=self.sig)
            self.save_variables(path+"/bart_initial_weights")

            from tensorflow.python.tools import saved_model_utils
            meta_graph_def = saved_model_utils.get_meta_graph_def(path, "serve")

            with open(path + "/bart_config.dat", 'w') as f:

                for signature in list(self.sig):

                    inputs = meta_graph_def.signature_def[signature].inputs
                    outputs = meta_graph_def.signature_def[signature].outputs

                    f.write('# ArgumentNameMapping\n')
                    f.write('{}\n'.format(signature))

                    for bart_name in list(inputs):
                        f.write("{} {} {}\n".format(bart_name, inputs[bart_name].name.split(":")[0], inputs[bart_name].name.split(":")[1]))

                    for bart_name in list(outputs):
                        f.write("{} {} {}\n".format(bart_name, outputs[bart_name].name.split(":")[0], outputs[bart_name].name.split(":")[1]))

    
    BartWrapper(model, dims).save(path)

def tf2_add_signature(path, signature='serving_default'):

	from tensorflow.python.tools import saved_model_utils
	meta_graph_def = saved_model_utils.get_meta_graph_def(path, "serve")

	with open(path + "/bart_config.dat", 'w') as f:
		
		inputs = meta_graph_def.signature_def[signature].inputs
		outputs = meta_graph_def.signature_def[signature].outputs

		f.write('# ArgumentNameMapping\n')
		f.write('{}\n'.format(signature))

		for bart_name in list(inputs):
			f.write("{} {} {}\n".format(bart_name, inputs[bart_name].name.split(":")[0], inputs[bart_name].name.split(":")[1]))

		for bart_name in list(outputs):
			f.write("{} {} {}\n".format(bart_name, outputs[bart_name].name.split(":")[0], outputs[bart_name].name.split(":")[1]))

class TensorMap:
    def __init__(self, tensor, name, enforce_real = False):

        if isinstance(tensor, TensorMap):
            self.tensor = tensor.tensor
        else:
            self.tensor = tensor
        self.name = name

        if (self.tensor.shape[-1] != 2) and (self.tensor.dtype == tf1.float32):
            self.type = "REAL"
        else:
            self.type = "COMPLEX"

        if isinstance(tensor, TensorMap):
            self.type = tensor.type

        if enforce_real:
            self.type = "REAL"

    def export(self):
        n = self.tensor.name
        return "{} {} {} {}".format(self.name, n.split(":")[0], n.split(":")[1], self.type)

def tf1_export_tensor_mapping(path, name, mapping, signature="serving_default"):
    with open(path + "/" + name + ".map", 'w') as f:
        f.write('# ArgumentNameMapping\n')
        f.write('{}\n'.format(signature))
        for map in mapping:
            f.write('{}\n'.format(map.export()))

def tf1_op_exists(graph, name):
    try:
        graph.get_operation_by_name(name)
        return True
    except KeyError:
        return False

def tf1_find_tensors(graph, inputs, outputs):
    
    if inputs is None:
        II = 0
        inputs = []
        while tf1_op_exists(graph, "input_"+str(II)):
            inputs.append(graph.get_tensor_by_name("input_{}:0".format(II)))
            II += 1

    if outputs is None:
        OO = 0
        outputs = []
        while tf1_op_exists(graph, "output_"+str(OO)):
            outputs.append(graph.get_tensor_by_name("output_{}:0".format(OO)))
            OO += 1

    for i in range(len(inputs)):
        inputs[i] = TensorMap(inputs[i], "input_"+str(i))

    for i in range(len(outputs)):
        outputs[i] = TensorMap(outputs[i], "output_"+str(i))

    return inputs, outputs


def tf1_graph_attach_gradients(graph, inputs, outputs):

    grad_tensors=[]

    for o, out in enumerate(outputs):
        with graph.as_default():
            gy = tf1.placeholder(out.tensor.dtype, shape=out.tensor.shape, name='grad_ys_'+ str(o))
            grad_tensors.append(TensorMap(gy, 'grad_ys_'+ str(o), out.type == "REAL"))

    for i, inp in enumerate(inputs):
        for o, out in enumerate(outputs):
            name = 'grad_{}_{}'.format(i, o)
            with graph.as_default():
                grad = tf1.gradients(out.tensor, inp.tensor, grad_tensors[o].tensor)
                grad = tf1.reshape(grad, tf1.shape(inp.tensor), name='grad_{}_{}'.format(i, o))
                grad_tensors.append(TensorMap(grad, name, inp.type == "REAL"))
    
    return grad_tensors


def tf1_export_graph(path, graph = None, session=None, inputs=None, outputs=None, name=None, attach_gradients=True):

    if graph is None:
        graph = tf1.get_default_graph()

    if name is None:
        name = os.path.basename(os.path.normpath(path))

    inputs, outputs = tf1_find_tensors(graph, inputs, outputs)

    mappings = []
    if attach_gradients:
        mappings = tf1_graph_attach_gradients(graph, inputs, outputs)

    mappings += inputs
    mappings += outputs

    tf1.train.write_graph(graph, path, name+'.pb', False)

    if session is not None:
        saver = tf1.train.Saver()
        saver.save(session, os.path.join(path, name))
    else:
        if (tf1_op_exists(graph, "save/restore_all")):
            print("WARNING: No weights are stored with the graph!\nWARNING: BART probably will not be able to load the graph.")

    tf1_export_tensor_mapping(path, name, mappings)


def tf1_convert_model(model_path, path, name):

    sess = tf1.Session()
    saver = tf1.train.Saver()
    saver.restore(sess, model_path) 

    tf1_graph_attach_gradients(sess.graph)
    tf1_export_graph(sess.graph, path, name, session=sess)



# Copyright 2013-2015. The Regents of the University of California.
# Copyright 2021. Uecker Lab. University Center Göttingen.
# Copyright 2024. Institute for Biomedical Imaging. TU Graz.
# All rights reserved. Use of this source code is governed by
# a BSD-style license which can be found in the LICENSE file.
#
# Authors:
# 2013 Martin Uecker <uecker@eecs.berkeley.edu>
# 2015 Jonathan Tamir <jtamir@eecs.berkeley.edu>


import numpy as np

# see src/misc/io.c for rawarray header definition
_RA_MAGIC = np.uint64(0x7961727261776172)
_RA_TYPE_COMPLEX = np.uint64(4)
_RA_CFL_SIZE = np.uint64(8)
_RA_FLAG_BIG_ENDIAN = np.uint64(1)
_RA_HEADER_ELEMS = 6


def _readra(name):
    with open(name, "rb") as f:
        magic, flags, eltype, elsize, datasize, ndims = np.fromfile(
            f, dtype=np.uint64, count=_RA_HEADER_ELEMS)

        if (magic != _RA_MAGIC
                or flags & _RA_FLAG_BIG_ENDIAN
                or eltype != _RA_TYPE_COMPLEX
                or elsize != _RA_CFL_SIZE):
            raise RuntimeError(f"Invalid .ra header: {name}")

        dims = np.fromfile(f, dtype=np.uint64, count=ndims)

        # remove trailing singleton dimensions
        while dims.size > 1 and dims[-1] == 1:
            dims = dims[:-1]

        return np.fromfile(f, dtype=np.complex64, count=int(datasize // 8)
                           ).reshape(dims, order='F')


def _writera(name, array):

    if array.dtype != np.complex64:
        array = array.astype(np.complex64)

    header = np.array([
        _RA_MAGIC,
        0,
        _RA_TYPE_COMPLEX,
        _RA_CFL_SIZE,
        array.nbytes,
        array.ndim
    ], dtype=np.uint64)
    shape_arr = np.array(array.shape, dtype=np.uint64)

    with open(name, "wb") as f:
        f.write(header)
        f.write(shape_arr)
        f.write(np.ascontiguousarray(array.T))


def readcfl(name):

    if name.endswith(".ra"):
        return _readra(name)

    # get dims from .hdr
    with open(name + ".hdr", "rt") as f:
        next(f)  # skip first line
        line = next(f)
    dims = [int(i) for i in line.split()]

    # remove singleton dimensions from the end
    while len(dims) > 1 and dims[-1] == 1:
        dims.pop()

    # load data and reshape into dims
    with open(name + ".cfl", "rb") as f:
        return np.fromfile(f, dtype=np.complex64, count=np.prod(dims)
                           ).reshape(dims, order='F')  # column-major


def readmulticfl(name):
    # get dims from .hdr
    with open(name + ".hdr", "rt") as h:
        lines = h.read().splitlines()

    index_dim = 1 + lines.index('# Dimensions')
    total_size = int(lines[index_dim])
    index_sizes = 1 + lines.index('# SizesDimensions')
    sizes = [int(i) for i in lines[index_sizes].split()]
    index_dims = 1 + lines.index('# MultiDimensions')

    with open(name + ".cfl", "rb") as d:
        a = np.fromfile(d, dtype=np.complex64, count=total_size)

    offset = 0
    result = []
    for i in range(len(sizes)):
        dims = ([int(i) for i in lines[index_dims + i].split()])
        n = np.prod(dims)
        result.append(a[offset:offset+n].reshape(dims, order='F'))
        offset += n

    if total_size != offset:
        print("Error")

    return result


def writecfl(name, array):

    if name.endswith(".ra"):
        return _writera(name, array)

    with open(name + ".hdr", "wt") as f:
        f.write('# Dimensions\n')
        f.write(" ".join(str(i) for i in array.shape))
        f.write('\n')

    if array.dtype != np.complex64:
        array = array.astype(np.complex64)

    with open(name + ".cfl", "wb") as f:
        f.write(np.ascontiguousarray(array.T))


def writemulticfl(name, arrays):
    total_size = sum(arr.size for arr in arrays)
    dims = [arr.shape for arr in arrays]

    with open(name + ".hdr", "wt") as f:
        f.write('# Dimensions\n')
        f.write("%d\n" % total_size)
        f.write('# SizesDimensions\n')
        f.write(' '.join(str(len(dim)) for dim in dims))
        f.write('\n')
        f.write('# MultiDimensions\n')
        for dim in dims:
            f.write(' '.join(str(i) for i in dim))
            f.write('\n')

    with open(name + ".cfl", "wb") as f:
        for array in arrays:
            if array.dtype != np.complex64:
                array = array.astype(np.complex64)
            f.write(np.ascontiguousarray(array.T))



#!/usr/bin/python3
#
# Copyright 2017. Massachusetts Institute of Technology.
# Copyright 2015. The Regents of the University of California.
# All rights reserved. Use of this source code is governed by
# a BSD-style license which can be found in the LICENSE file.
#
# Authors:
# 2015 Frank Ong <frankong@berkeley.edu>
# 2017 Siddharth Iyer <ssi@mit.edu>

import operator
import numpy as np
import sys
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons
from functools import partial, reduce
import time
import threading
import os.path

class DiscreteSlider(Slider):
    """A matplotlib slider widget with discrete steps."""
    def __init__(self, *args, **kwargs):
        self.previous_val = kwargs['valinit']
        Slider.__init__(self, *args, **kwargs)

    def set_val(self, val):
        discrete_val = round(val)
        xy = self.poly.xy
        xy[2] = discrete_val, 1
        xy[3] = discrete_val, 0
        self.poly.xy = xy
        self.valtext.set_text(self.valfmt % discrete_val)
        if self.drawon: 
            self.ax.figure.canvas.draw()
        self.val = val
        if self.previous_val!=discrete_val:
            self.previous_val = discrete_val
            if not self.eventson: 
                return
            for cid, func in self.observers.items():
                func(discrete_val)


class BartView(object):

    def __init__(self, cflname):
        
        matplotlib.rcParams['toolbar'] = 'None'
        #matplotlib.rcParams['font.size'] = 6

        # Read data
        self.cflname = sys.argv[1]
        self.im = self.readcfl(self.cflname)
        self.im_unsqueeze_shape = np.where( np.array(self.im.shape) > 1 )[0]
        self.im = self.im.squeeze()
        if sys.version_info.major==3 and sys.version_info.minor < 8:
           t1 = time.clock()

        # Reorder image
        self.Ndims = len( self.im.shape )
        self.order = np.r_[:self.Ndims]
        self.im_ordered = self.im
        self.order3 = np.array([0,1,1])
        
        # Slice image
        self.slice_num = np.zeros( self.Ndims, dtype = 'int' );
        self.im_shape = self.im_ordered.shape
        self.im_slice = self.im_ordered[ (slice(None), slice(None)) + tuple(self.slice_num[2:]) ]
        
        # Create figure
        self.fig = plt.figure(facecolor='black', figsize=(9,6))
        #self.fig = plt.figure(facecolor='black', figsize=(6,4))
        self.fig.subplots_adjust( left=0.0 , bottom=0.0 , right=1.0 , top=1 - 0.25)
        self.fig.canvas.set_window_title(self.cflname)
        
        # Show image
        self.immax = np.max(abs(self.im))
        self.l = plt.imshow( abs(self.im_slice) , cmap = "gray", vmin=0, vmax=self.immax)
        self.ax = plt.gca()
        self.asp = self.im_ordered.shape[1] / self.im_ordered.shape[0]
        self.aspect = 1
        self.ax.set_aspect( 1 )
        plt.axis('off')
        
        radios = []
        buttons = []
        sliders = []
        

        # Create Radio Buttons for X Y dimensions
        dims = self.im_unsqueeze_shape[ self.order ].astype(str)
        for i in range(0,len(dims)):
            dims[i] = "Dim " + dims[i]
        oboxx_ax = plt.axes( [0, 1 - 0.03, 0.1, 0.03], facecolor = "gainsboro" )
        oboxx_ax.set_xticks([]);
        oboxx_ax.set_yticks([]);
        orderx_ax = plt.axes( [0, 1 - 0.18, 0.1, 0.15], facecolor = 'gainsboro' )
        orderx_radio = RadioButtons( orderx_ax, dims, activecolor = 'SteelBlue', active = 0 )
        orderx_ax.text(0.5,1.05,  'Up/Down', horizontalalignment = 'center')
        radios.append( orderx_radio )
        orderx_radio.on_clicked( self.update_orderx )

        oboxy_ax = plt.axes( [0.1, 1 - 0.03, 0.1, 0.03], facecolor = "gainsboro" )
        oboxy_ax.set_xticks([]);
        oboxy_ax.set_yticks([]);
        ordery_ax = plt.axes( [0.1, 1 - 0.18, 0.1, 0.15], facecolor = 'gainsboro' )
        ordery_radio = RadioButtons( ordery_ax, dims, activecolor = 'SteelBlue', active = 1 )
        ordery_ax.text(0.5,1.05,  'Left/Right', horizontalalignment = 'center')
        radios.append( ordery_radio )
        ordery_radio.on_clicked( self.update_ordery )
    

        # Create Radio buttons for mosaic
        self.mosaic_valid = False
        mbox_ax = plt.axes( [0.2, 1 - 0.03, 0.1, 0.03], facecolor = "gainsboro" )
        mbox_ax.set_xticks([]);
        mbox_ax.set_yticks([]);
        mosaic_ax = plt.axes( [0.2, 1 - 0.18, 0.1, 0.15], facecolor = 'gainsboro' )
        mosaic_radio = RadioButtons( mosaic_ax, dims, activecolor = 'SteelBlue', active = 1 )
        mosaic_ax.text(0.5,1.05,  'Mosaic', horizontalalignment = 'center')
        radios.append( mosaic_radio )
        mosaic_radio.on_clicked( self.update_mosaic )
    

            
        # Create flip buttons
        self.flipx = 1;
        flipx_ax = plt.axes( [0.3, 1 - 0.09, 0.1, 0.09] )
        flipx_button = Button( flipx_ax, 'Flip\nUp/Down', color='gainsboro' )
        flipx_button.on_clicked(self.update_flipx);

        self.flipy = 1;
        flipy_ax = plt.axes( [0.3, 1 - 0.18, 0.1, 0.09] )
        flipy_button = Button( flipy_ax, 'Flip\nLeft/Right', color='gainsboro' )
        flipy_button.on_clicked(self.update_flipy);

        
        # Create Refresh buttons
        refresh_ax = plt.axes( [0.4, 1 - 0.09, 0.1, 0.09] )
        refresh_button = Button( refresh_ax, 'Refresh', color='gainsboro' )
        refresh_button.on_clicked(self.update_refresh);

        # Create Save button
        
        save_ax = plt.axes( [0.4, 1 - 0.18, 0.1, 0.09] )
        save_button = Button( save_ax, 'Export to\nPNG', color='gainsboro' )
        save_button.on_clicked(self.save);


        # Create dynamic refresh radio button
        #self.drefresh = threading.Event()
        #drefresh_ax = plt.axes( [0.4, 1 - 0.18, 0.1, 0.09] )
        #drefresh_button = Button( drefresh_ax, 'Dynamic\nRefresh', color='gainsboro' )
        #drefresh_button.on_clicked(self.update_drefresh);


        # Create Magnitude/phase radio button
        self.magnitude = True
        mag_ax = plt.axes( [0.5, 1 - 0.18, 0.1, 0.18], facecolor = 'gainsboro' )
        mag_radio = RadioButtons( mag_ax, ('Mag','Phase') , activecolor = 'SteelBlue', active = 0 )
        radios.append( mag_radio )
        mag_radio.on_clicked( self.update_magnitude )

        
        sbox_ax = plt.axes( [0.6, 1 - 0.18, 0.5, 0.18], facecolor='gainsboro')
        sbox_ax.set_xticks([])
        sbox_ax.set_yticks([])
        
        # Create aspect sliders
        aspect_ax = plt.axes( [0.65, 1 - 0.09 + 0.02, 0.1, 0.02], facecolor = 'white' )
        aspect_slider = Slider( aspect_ax, "", 0.25, 4, valinit=1, color='SteelBlue')
        aspect_ax.text( 4 / 2,1.5,  'Aspect Ratio', horizontalalignment = 'center')
        sliders.append( aspect_slider )
        aspect_slider.on_changed( self.update_aspect )
        
        # Create contrast sliders
        self.vmin = 0
        vmin_ax = plt.axes( [0.83, 1 - 0.09 + 0.02, 0.1, 0.02], facecolor = 'white' )
        vmin_slider = Slider( vmin_ax, "", 0, 1, valinit=0, color='SteelBlue')
        vmin_ax.text(0.5,1.5,  'Contrast Min', horizontalalignment = 'center')
        sliders.append( vmin_slider )
        vmin_slider.on_changed( self.update_vmin )

        self.vmax = 1
        vmax_ax = plt.axes( [0.83, 1 - 0.18 + 0.02, 0.1, 0.02], facecolor = 'white' )
        vmax_slider = Slider( vmax_ax, "", 0, 1, valinit=1, color='SteelBlue')
        vmax_ax.text(0.5,1.5,  'Contrast Max', horizontalalignment = 'center')
        sliders.append( vmax_slider )
        vmax_slider.on_changed( self.update_vmax )

        
        # Create sliders for choosing slices
        box_ax = plt.axes( [0, 1 - 0.25, 1, 0.07], facecolor='gainsboro')
        box_ax.set_xticks([])
        box_ax.set_yticks([])

        slider_thick = 0.02
        slider_start = 0.1
        ax = []
        for d in np.r_[:self.Ndims]:
            slice_ax = plt.axes( [0.01 + 1 / self.Ndims * d, 1 - 0.24, 0.8 / self.Ndims, slider_thick] , facecolor='white')
            slice_slider = DiscreteSlider( slice_ax, "", 0, self.im_shape[d]-1, valinit=self.slice_num[d],valfmt='%i', color='SteelBlue')
            slice_ax.text( (self.im_shape[d]-1)/2,1.5,  'Dim %d Slice' % self.im_unsqueeze_shape[d], horizontalalignment = 'center' )
            sliders.append(slice_slider);
            slice_slider.on_changed( partial( self.update_slice, d ) )
            

        plt.show()

    def readcfl(self, name):
            h = open(name + ".hdr", "r")
            h.readline() # skip
            l = h.readline()
            dims = [int(i) for i in l.split( )]
            n = reduce(operator.mul, dims, 1)
            h.close()
            return np.memmap( name + ".cfl", dtype = np.complex64, mode='r', shape=tuple(dims), order='F' )

    def save( self, event ):
        extent = self.ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())
        num = 0
        fname = self.cflname + '_' + str(num) + '.png'
        while( os.path.isfile(fname)  ):
            num += 1
            fname = self.cflname + '_' + str(num) + '.png'
            
        self.fig.savefig(fname, bbox_inches=extent)

    def update_flipx( self, event ):
        self.flipx *= -1
        self.update_image()

        
    def update_flipy( self, event ):
        self.flipy *= -1
        self.update_image()

    def update_refresh( self, event ):
        self.update_image()

    def dynamic_refresh( self ):
        while( self.drefresh.is_set() ):
            self.update_image()

    def update_drefresh( self, event ):
        if ( self.drefresh.is_set() ):
            self.drefresh.clear()
        else:
            self.drefresh.set()
            th = threading.Thread( target = self.dynamic_refresh )
            th.start()
            
    def update_aspect( self, aspect ):
        self.aspect = aspect
        self.ax.set_aspect( self.asp * self.im_ordered.shape[0] / self.im_ordered.shape[1] * aspect )

    def update_vmax( self, vmax ):
        self.vmax = max(self.vmin, vmax)
        self.l.set_clim( vmax = self.vmax * self.immax );

    def update_vmin( self, vmin ):
        self.vmin = min(self.vmax,vmin)
        self.l.set_clim( vmin = self.vmin * self.immax );

    def update_magnitude( self, l ):
        self.magnitude =  ( l == 'Mag' )
        if (self.magnitude):
            self.l.set_cmap('gray')
        else:
            self.l.set_cmap('hsv')
        self.update_image()
            
    def update_orderx( self, l ):
        l = int(l[4:])
        self.order3[0] = np.where( self.im_unsqueeze_shape == l )[0]
        self.update_ordered_image()

    def update_ordery( self, l ):
        l = int(l[4:])
        self.order3[1] = np.where( self.im_unsqueeze_shape == l )[0]
        self.update_ordered_image()

        
    def update_ordered_image(self):
        self.mosaic_valid = len( self.order3[:3] ) == len( set( self.order3[:3] ) )
        self.order_valid = len( self.order3[:2] ) == len( set( self.order3[:2] ) );
        
        if ( self.mosaic_valid ):
            self.order[:3] = self.order3[:3]
            order_remain = np.r_[:self.Ndims]
            for t in np.r_[:3]:
                order_remain = order_remain[  (order_remain != self.order[t] ) ]
            self.order[3:] = order_remain
            self.im_ordered = np.transpose( self.im, self.order )
            self.ax.set_aspect( self.asp * self.im_ordered.shape[0] / self.im_ordered.shape[1] * self.aspect )
            self.update_image()
        elif ( self.order_valid ):
            self.order[:2] = self.order3[:2]
            order_remain = np.r_[:self.Ndims]
            for t in np.r_[:2]:
                order_remain = order_remain[  (order_remain != self.order[t] ) ]
            self.order[2:] = order_remain
            self.im_ordered = np.transpose( self.im, self.order )
            self.ax.set_aspect( self.asp * self.im_ordered.shape[0] / self.im_ordered.shape[1] * self.aspect )
            self.update_image()


    def update_image( self ):
        self.immax = np.max(abs(self.im))
        self.l.set_clim( vmin = self.vmin * self.immax ,  vmax = self.vmax * self.immax );
        if ( self.mosaic_valid ):
            im_slice = self.im_ordered[ (slice(None,None,self.flipx), slice(None,None,self.flipy), slice(None)) + tuple(self.slice_num[self.order[3:]])]
            im_slice = self.mosaic( im_slice )
        else:
            im_slice = self.im_ordered[ (slice(None,None,self.flipx), slice(None,None,self.flipy)) + tuple(self.slice_num[self.order[2:]]) ]
        
        if self.magnitude:
            self.l.set_data( abs(im_slice) )
        else:
            self.l.set_data( (np.angle(im_slice) + np.pi) / (2 * np.pi) )


        
        self.fig.canvas.draw()

    def update_slice( self, d, s ):
        self.slice_num[d] = int(round(s))
        self.update_image()

    def mosaic( self, im ):
        im = im.squeeze()
        (x, y, z) = im.shape
        z2 = int( np.ceil( z ** 0.5 ) )
        z = int( z2 ** 2 )
        im = np.pad( im, [(0,0), (0,0), (0, z - im.shape[2] )], mode='constant')
        im = im.reshape(  (x, y * z, 1), order = 'F' )

        im = im.transpose( (1, 2, 0) )
        im = im.reshape( (y * z2 , z2, x), order = 'F' )
        im = im.transpose( (2, 1, 0) )
        im = im.reshape( (x * z2, y * z2), order = 'F' )

        return im

    def update_mosaic( self, l ):
        l = int(l[4:])
        self.order3[2] = np.where( self.im_unsqueeze_shape == l )[0]
        self.update_ordered_image()

if __name__ == "__main__":

    # Error if more than 1 argument
    if (len(sys.argv) != 2):
        print("BartView: multidimensional image viewer for cfl")
        print("Usage: bview cflname")
        exit()

    BartView( sys.argv[1] )

    



#!/usr/bin/env python3
import os
import re

def parse_hdr(fname):
    ret = {}
    key = None
    with open(fname, 'r') as f:
        for line in f:
            if line.startswith("# "):
                key = line[2:].strip()
            elif not key is None:
                ret[key] = line.strip()
    return ret

def short_file_name(fname):
    fname = fname.split("/")[-1]
    ending = fname.split(".")[-1]
    if ending in ["hdr"]:
        cut = -1 * (len(ending) + 1)
        fname = fname[:cut]
    return fname


start_level = 1
colorscheme_max_level = 7

def level_dot(level):
    return f"fillcolor={level if level <= colorscheme_max_level else 'gray'},style=filled"

def File_node(tool_node, hdr, level=start_level):
    name = short_file_name(hdr)
    node_id = f"{name}_{tool_node['id'] if tool_node else 'ROOT'}"
    ftype = 'pipe' if hdr.endswith(".fifo") or hdr == '-' else 'file'
    return {'id': node_id, 'type': ftype, 'level': level,
            'dot': f" [label=\"{name}\",shape=diamond,{level_dot(level)}]" }

def Tool_node(node_id, level=start_level):
    if node_id is None:
        return None
    name = node_id.split("_")[0]
    d = { 'id': node_id, 'type': 'tool','src': [], 'level': level, 'sync' : False,
         'dot': f" [label=\"{name}\",{level_dot(level)}]" }
    try:
        info = parse_hdr(os.path.join(node_dir, node_id))
        for src in info.get("Input Nodes", "").split():
            x = src.split(":")
            d['src'].append((x[0], x[1], x[1].endswith(".fifo") or x[1] == '-'))
        d['command'] = info.get("Command")
        d['directory'] = info.get("Directory")
    except FileNotFoundError:
        d['deleted'] = True
    return d

def Edge(node, src, name=None):
    key = f"{src['id']}->{node['id']}"
    label = ""
    if name and name != "-":
        label = f" [label=\"{short_file_name(name)}\"]"
    return { key:  {'dot': key + label } }


class Graph:
    def __init__(self, start_file, node_dir):
        self.edges = {}
        self.nodes = {}
        self.src_nodes = {}
        self.desc_nodes = {}

        self._make_graph(start_file, node_dir)

    def add_edge(self, node, src, name=None):
        self.edges.update(Edge(node, src, name))
        if not node['id'] in self.src_nodes:
            self.src_nodes[node['id']] = {}
        if not src['id'] in self.desc_nodes:
            self.desc_nodes[src['id']] = {}
        self.src_nodes[node['id']][src['id']] = name
        self.desc_nodes[src['id']][node['id']] = name

    def add_node(self, node):
        self.nodes[node['id']] = node

    def node_has_stdin(self, node):
        return "-" in self.src_nodes.get(node['id'], {}).values()

    def node_has_stdout(self, node):
        return "-" in self.desc_nodes.get(node['id'], {}).values()

    def _step(graph, node):
        graph.add_node(node)
        src_nodes = []
        for src_node_id, src_name, src_is_piped in node['src']:
            level = node['level'] + (0 if src_is_piped else 1)
            src_node = Tool_node(src_node_id, level)
            if src_is_piped:
                graph.add_edge(node, src_node, src_name)
            else:
                src_node['sync'] = True
                file_node = File_node(src_node, src_name, level)
                graph.add_node(file_node)
                graph.add_edge(file_node, src_node)
                graph.add_edge(node, file_node)

            if src_node_id not in graph.nodes or level > graph.nodes[src_node_id]['level']:
                src_nodes.append(src_node)

        return src_nodes

    def _make_graph(graph, start_file, node_dir):
        graph.start_node = Tool_node(parse_hdr(start_file).get("Node-ID"))
        graph.start_node['sync'] = True
        file_node = File_node(graph.start_node, start_file)

        graph.add_node(file_node)
        if graph.start_node:
            graph.add_edge(file_node, graph.start_node)
        src_nodes = [graph.start_node] if graph.start_node else []

        while len(src_nodes) > 0:
            new_src_nodes = []
            for node in src_nodes:
                new_src_nodes += graph._step(node)
            src_nodes = new_src_nodes

    def get_dot(graph):
        ret = "digraph {\n\tnode [colorscheme=" + f"greens{colorscheme_max_level}" + "];\n"
        for node_id, node_opts in graph.nodes.items():
            ret += f"\t{node_id}{node_opts['dot']};\n"
        for edge in graph.edges.values():
            ret += f"\t{edge['dot']};\n"
        ret += "}\n"
        return ret

    def get_heads(self):
        heads = set()
        srcs = self.src_nodes[self.start_node['id']]
        while len(srcs) > 0:
            new_srcs = {}
            for src in srcs:
                x = self.src_nodes.get(src, {})
                if len(x) == 0:
                    heads.add(src)
                new_srcs.update(x)
            srcs = new_srcs
        return list(heads)

    def create_cmds(self, filter_path = None):
        heads = self.get_heads()
        max_level = 1
        for node_id in heads:
            max_level = max(max_level, self.nodes[node_id]['level'])

        level = max_level
        cmds = []
        done = {}
        while level > 0:
            new_cmds = []
            updates = 1
            strands = []
            while updates:
                updates = 0
                new_heads = []
                for node_id in heads:
                    strand = []
                    while node_id:
                        node = self.nodes[node_id]
                        desc = list(self.desc_nodes.get(node['id'], {}).keys())

                        if node['type'] == 'file':
                            new_heads += desc
                            break;
                        if node_id in done:
                            break;
                        if node['level'] != level:
                            new_heads += [node_id]
                            break

                        updates += 1
                        strand.append(node)
                        done[node_id] = 1

                        node_id = None

                        if len(desc) == 1 and not (self.node_has_stdin(self.nodes[desc[0]]) and (self.src_nodes[desc[0]][node['id']] != "-")):
                            node_id = desc[0]
                        else:
                            for desc_id, name in self.desc_nodes[node['id']].items():
                                # continue along stdout
                                if name == "-":
                                    node_id = desc_id
                                else:
                                    new_heads.append(desc_id)
                    strands.append(strand)

                heads = new_heads
            for strand in sorted(strands, key = lambda x: len(x) > 0 and x[-1]['sync']):
                cmds.append(strand)
            level -= 1

        if filter_path:
            rex = re.compile(filter_path)

        ret = "#!/bin/bash\nexport BASEDIR=$(realpath .)\n"
        d = ""
        last_d = None
        e = ""
        t = ""
        for strand in cmds:
            for i, cmd in enumerate(strand):
                last_d = d
                d = cmd['directory']
                if filter_path:
                    d = rex.sub('', d)
                if last_d != d:
                    t = '\t'
                    ret += f"{e}(\n{t}mkdir -p {d}; cd {d};\n"
                    e = ");\n"

                if cmd['sync']:
                    sep = ";"
                elif self.node_has_stdout(cmd):
                    if i + 1 < len(strand) and strand[i + 1]['directory'] == d:
                        sep = " |\\"
                    else:
                        sep = " &"
                        e = ") |\\\n"
                else:
                    sep = " &"

                c = cmd['command']
                if filter_path:
                    c = rex.sub('$BASEDIR/', c)
                ret += t + c + sep  + '\n'
        return ret + e

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("infile", help=".hdr-file for which to generate graph")
    parser.add_argument("--type", choices=['dot', 'bash'], default='dot', help="Output file type")
    parser.add_argument("--filter_path", help="Remove path prefix for script output")
    args = parser.parse_args()

    node_dir = os.environ["BART_TOOL_GRAPH"]
    graph = Graph(args.infile, node_dir)

    if args.type == 'dot':
        print(graph.get_dot())
    elif args.type == 'bash':
        print(graph.create_cmds(args.filter_path))
